
local PLAYER = 'player'
local LEGENDARY = 5
local ARTIFACT = 6

local CAST_EVENTS = {
  'UNIT_SPELLCAST_CHANNEL_STOP',
  'UNIT_SPELLCAST_DELAYED',
  'UNIT_SPELLCAST_FAILED',
  'UNIT_SPELLCAST_FAILED_QUIET',
  'UNIT_SPELLCAST_INTERRUPTED',
  'UNIT_SPELLCAST_STOP',
  'UNIT_SPELLCAST_SUCCEEDED',
}

local RESTORE_EVENTS = {
  'PLAYER_ALIVE',
  'PLAYER_CAMPING',
  'PLAYER_CONTROL_GAINED',
  'PLAYER_ENTERING_BATTLEGROUND',
  'PLAYER_ENTERING_WORLD',
  'PLAYER_LEAVE_COMBAT',
  'PLAYER_LEAVING_WORLD',
  'PLAYER_LOGIN',
  'PLAYER_LOGOUT',
  'PLAYER_REGEN_ENABLED',
  'PLAYER_STARTED_MOVING',
  'PLAYER_STOPPED_MOVING',
  'PLAYER_UNGHOST',
  'PLAYER_UPDATE_RESTING',
}

local function EquipBlocked()
  return InCombatLockdown() or UnitCastingInfo(PLAYER)
end

local function GetItem(slot)
  return GetInventoryItemID(PLAYER, slot)
end

local function GetSpellID(spell)
  return select(7, GetSpellInfo(spell))
end

local function GetRarity(item)
  if item then
    local rarity = select(3, GetItemInfo(item))
    return rarity
  end
end

local function IsArtifact(item)
  return GetRarity(item) == ARTIFACT
end

local FC = CreateExecutor()

function FC.Reset(self)
  self.items = {}
  self.set = {}
  self.spell_id = nil
  self:UnregisterAllEvents()
end

function FC.OnCast(self, event, unit, _, spell_id)
  if self.spell_id and unit == PLAYER and self.spell_id == spell_id then
    self.spell_id = nil
  end
  if not self.spell_id then
    self:Restore()
  end
end

function FC.OnEquip(self, event, slot)
  local item = GetItem(slot)
  if next(self.set) then
    if self.set[slot] == item then
      self.set[slot] = nil
    end
    if not next(self.set) then
      self:Restore()
    end
  elseif next(self.items) then
    if self.items[slot] == item then
      self.items[slot] = nil
    end
    if not next(self.items) then
      self:Reset()
    end
  end
end

function FC.Restore(self)
  if next(self.set) or self.spell_id or EquipBlocked() then
    return
  end
  for slot, item in pairs(self.items) do
    if item == GetItem(slot) then
      self.items[slot] = nil
    elseif GetRarity(item) == LEGENDARY then
      self:RunAfter(0.5, EquipItemByName, item, slot)
    else
      EquipItemByName(item, slot)
    end
  end
  if not next(self.items) then
    self:Reset()
  end
end

function FC.Run(self, set, spell)
  if next(self.items) or EquipBlocked() then
    return
  end
  local set_id = C_EquipmentSet.GetEquipmentSetID(set)
  if not set_id then
    print('FastCast Error: Invalid equipment set "' .. set .. '"')
    self.set = nil
    return
  end
  self.spell_id = GetSpellID(spell)
  if not self.spell_id then
    print('FastCast Error: Invalid spell "' .. spell .. '"')
    print('  For casting a toy or item use the numeric spellID ')
    print('  instead of the casting name (find it on Wowhead).')
    print('  For example, use 253937 instead of "Flight Master\'s Whistle"')
    print('  and 147420 instead of "One with Nature" for Nature\'s Beacon.')
    print('  This is a new limitation of the 8.0 interface changes.')
    return
  end
  self.set = C_EquipmentSet.GetItemIDs(set_id)
  for slot = INVSLOT_FIRST_EQUIPPED, INVSLOT_LAST_EQUIPPED do
    local item = GetItem(slot)
    local set_item = self.set[slot]
    local swap = true
    if not set_item then
      swap = (slot == INVSLOT_OFFHAND and not IsArtifact(set_item))
    elseif item == set_item or set_item <= 0 then
      swap = false
    end
    if not swap then
      item = nil
      set_item = nil
    end
    self.items[slot] = item
    self.set[slot] = set_item
  end
  if not next(self.items) then
    return
  end
  self:RegisterEvents()
  C_EquipmentSet.UseEquipmentSet(set_id)
end

function FC.Init(self)
  self:Reset()
  self:AddEvent('PLAYER_EQUIPMENT_CHANGED', self.OnEquip)
  for _, event in ipairs(CAST_EVENTS) do
    self:AddEvent(event, self.OnCast)
  end
  for _, event in ipairs(RESTORE_EVENTS) do
    self:AddEvent(event, self.Restore)
  end
end

FC:Init()

function FastCast(...)
  FC:Run(...)
end
